"""Parse mystery style generated by MTA at caiwireless.net."""

import re

from email.iterators import body_line_iterator
from enum import Enum
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


tcre = re.compile(
    r'the following recipients did not receive this message:',
    re.IGNORECASE)
acre = re.compile(
    r'<(?P<addr>[^>]*)>')


class ParseState(Enum):
    start = 0
    tag_seen = 1


@public
@implementer(IBounceDetector)
class Caiwireless:
    """Parse mystery style generated by MTA at caiwireless.net."""

    def process(self, msg):
        if msg.get_content_type() == 'multipart/mixed':
            state = ParseState.start
            # This format thinks it's a MIME, but it really isn't.
            for line in body_line_iterator(msg):
                line = line.strip()
                if state is ParseState.start and tcre.match(line):
                    state = ParseState.tag_seen
                elif state is ParseState.tag_seen and line:
                    mo = acre.match(line)
                    if mo:
                        return NoTemporaryFailures, set(mo.group('addr'))
                    else:
                        break
        return NoFailures



"""Interfaces."""

from public import public
from zope.interface import Interface


# Constants for improved readability in detector classes.  Use these like so:
#
# - to signal that no temporary or permanent failures were found:
#   `return NoFailures`
# - to signal that no temporary failures, but some permanent failures were
#   found:
#   `return NoTemporaryFailures, my_permanent_failures`
# - to signal that some temporary failures, but no permanent failures were
#   found:
#   `return my_temporary_failures, NoPermanentFailures`

NoTemporaryFailures = NoPermanentFailures = ()
NoFailures = (NoTemporaryFailures, NoPermanentFailures)

public(NoTemporaryFailures=NoTemporaryFailures)
public(NoPermanentFailures=NoPermanentFailures)
public(NoFailures=NoFailures)


@public
class IBounceDetector(Interface):
    """Detect a bounce in an email message."""

    def process(self, msg):
        """Scan an email message looking for bounce addresses.

        :param msg: An email message.
        :type msg: `Message`
        :return: A 2-tuple of the detected temporary and permanent bouncing
            addresses.  Both elements of the tuple are sets of string
            email addresses.  Not all detectors can tell the difference
            between temporary and permanent failures, in which case, the
            addresses will be considered to be permanently bouncing.
        :rtype: (set of strings, set of string)
        """



"""sina.com bounces"""

import re

from email.iterators import body_line_iterator
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


acre = re.compile(r'<(?P<addr>[^>]*)>')


@public
@implementer(IBounceDetector)
class Sina:
    """sina.com bounces"""

    def process(self, msg):
        """See `IBounceDetector`."""
        if msg.get('from', '').lower() != 'mailer-daemon@sina.com':
            return NoFailures
        if not msg.is_multipart():
            return NoFailures
        # The interesting bits are in the first text/plain multipart.
        part = None
        try:
            part = msg.get_payload(0)
        except IndexError:
            pass
        if not part:
            return NoFailures
        addresses = set()
        for line in body_line_iterator(part):
            mo = acre.match(line)
            if mo:
                addresses.add(mo.group('addr').encode('us-ascii'))
        return NoTemporaryFailures, addresses



"""Recognizes simple heuristically delimited bounces."""

import re

from email.iterators import body_line_iterator
from email.quoprimime import unquote
from enum import Enum
from flufl.bounce.interfaces import IBounceDetector, NoTemporaryFailures
from public import public
from zope.interface import implementer


class ParseState(Enum):
    start = 0
    tag_seen = 1


def _unquote_match(match):
    return unquote(match.group(0))


def _quopri_decode(address):
    # Some addresses come back with quopri encoded spaces.  This will decode
    # them and strip the spaces.  We can't use the undocumebted
    # email.quoprimime.header_decode() because that also turns underscores
    # into spaces, which is not good for us.  Instead we'll use the
    # undocumented email.quoprimime.unquote().
    #
    # For compatibility with Python 3, the API requires byte addresses.
    unquoted = re.sub('=[a-fA-F0-9]{2}', _unquote_match, address)
    try:
        return unquoted.encode('us-ascii').strip()
    except UnicodeEncodeError:
        return b''


def _c(pattern):
    return re.compile(pattern, re.IGNORECASE)


# Pattern to match any valid email address and not much more.
VALID = _c(rb'^[\x21-\x3d\x3f\x41-\x7e]+@[a-z0-9._]+$')


# This is a list of tuples of the form
#
#     (start cre, end cre, address cre)
#
# where 'cre' means compiled regular expression, start is the line just before
# the bouncing address block, end is the line just after the bouncing address
# block, and address cre is the regexp that will recognize the addresses.  It
# must have a group called 'addr' which will contain exactly and only the
# address that bounced.
PATTERNS = [
    # sdm.de
    (_c(r'here is your list of failed recipients'),
     _c(r'here is your returned mail'),
     _c(r'<(?P<addr>[^>]*)>')),
    # sz-sb.de, corridor.com, nfg.nl
    (_c(r'the following addresses had'),
     _c(r'transcript of session follows'),
     _c(r'^ *(\(expanded from: )?<?(?P<addr>[^\s@]+@[^\s@>]+?)>?\)?\s*$')),
    # robanal.demon.co.uk
    (_c(r'this message was created automatically by mail delivery software'),
     _c(r'original message follows'),
     _c(r'rcpt to:\s*<(?P<addr>[^>]*)>')),
    # s1.com (InterScan E-Mail VirusWall NT ???)
    (_c(r'message from interscan e-mail viruswall nt'),
     _c(r'end of message'),
     _c(r'rcpt to:\s*<(?P<addr>[^>]*)>')),
    # Smail
    (_c(r'failed addresses follow:'),
     _c(r'message text follows:'),
     _c(r'\s*(?P<addr>\S+@\S+)')),
    # newmail.ru
    (_c(r'This is the machine generated message from mail service.'),
     _c(r'--- Below the next line is a copy of the message.'),
     _c(r'<(?P<addr>[^>]*)>')),
    # turbosport.com runs something called `MDaemon 3.5.2' ???
    (_c(r'The following addresses did NOT receive a copy of your message:'),
     _c(r'--- Session Transcript ---'),
     _c(r'[>]\s*(?P<addr>.*)$')),
    # usa.net
    (_c(r'Intended recipient:\s*(?P<addr>.*)$'),
     _c(r'--------RETURNED MAIL FOLLOWS--------'),
     _c(r'Intended recipient:\s*(?P<addr>.*)$')),
    # hotpop.com
    (_c(r'Undeliverable Address:\s*(?P<addr>.*)$'),
     _c(r'Original message attached'),
     _c(r'Undeliverable Address:\s*(?P<addr>.*)$')),
    # Another demon.co.uk format
    (_c(r'This message was created automatically by mail delivery'),
     _c(r'^---- START OF RETURNED MESSAGE ----'),
     _c(r"addressed to '(?P<addr>[^']*)'")),
    # Prodigy.net full mailbox
    (_c("User's mailbox is full:"),
     _c(r'Unable to deliver mail.'),
     _c(r"User's mailbox is full:\s*<(?P<addr>[^>]*)>")),
    # Microsoft SMTPSVC
    (_c(r'The email below could not be delivered to the following user:'),
     _c(r'Old message:'),
     _c(r'<(?P<addr>[^>]*)>')),
    # Yahoo on behalf of other domains like sbcglobal.net
    (_c(r'Unable to deliver message to the following address\(es\)\.'),
     _c(r'--- Original message follows\.'),
     _c(r'<(?P<addr>[^>]*)>:')),
    # googlemail.com
    (_c(r'Delivery to the following recipient failed'),
     _c(r'----- Original message -----'),
     _c(r'^\s*(?P<addr>[^\s@]+@[^\s@]+)\s*$')),
    # kundenserver.de
    (_c(r'A message that you sent could not be delivered'),
     _c(r'^---'),
     _c(r'<(?P<addr>[^>]*)>')),
    # another kundenserver.de
    (_c(r'A message that you sent could not be delivered'),
     _c(r'^---'),
     _c(r'^(?P<addr>[^\s@]+@[^\s@:]+):')),
    # thehartford.com / songbird
    (_c(r'Del(i|e)very to the following recipients (failed|was aborted)'),
     # this one may or may not have the original message, but there's nothing
     # unique to stop on, so stop on the first line of at least 3 characters
     # that doesn't start with 'D' (to not stop immediately) and has no '@'.
     # Also note that simple_30.txt contains an apparent misspelling in the
     # MTA's DSN section.
     _c(r'^[^D][^@]{2,}$'),
     _c(r'^[\s*]*(?P<addr>[^\s@]+@[^\s@]+)\s*$')),
    # and another thehartfod.com/hartfordlife.com
    (_c(r'^Your message\s*$'),
     _c(r'^because:'),
     _c(r'^\s*(?P<addr>[^\s@]+@[^\s@]+)\s*$')),
    # kviv.be (InterScan NT)
    (_c(r'^Unable to deliver message to'),
     _c(r'\*+\s+End of message\s+\*+'),
     _c(r'<(?P<addr>[^>]*)>')),
    # earthlink.net supported domains
    (_c(r'^Sorry, unable to deliver your message to'),
     _c(r'^A copy of the original message'),
     _c(r'\s*(?P<addr>[^\s@]+@[^\s@]+)\s+')),
    # ademe.fr
    (_c(r'^A message could not be delivered to:'),
     _c(r'^Subject:'),
     _c(r'^\s*(?P<addr>[^\s@]+@[^\s@]+)\s*$')),
    # andrew.ac.jp
    (_c(r'^Invalid final delivery userid:'),
     _c(r'^Original message follows.'),
     _c(r'\s*(?P<addr>[^\s@]+@[^\s@]+)\s*$')),
    # E500_SMTP_Mail_Service@lerctr.org
    (_c(r'------ Failed Recipients ------'),
     _c(r'-------- Returned Mail --------'),
     _c(r'<(?P<addr>[^>]*)>')),
    # cynergycom.net
    (_c(r'A message that you sent could not be delivered'),
     _c(r'^---'),
     _c(r'(?P<addr>[^\s@]+@[^\s@)]+)')),
    # LSMTP for Windows
    (_c(r'^--> Error description:\s*$'),
     _c(r'^Error-End:'),
     _c(r'^Error-for:\s+(?P<addr>[^\s@]+@[^\s@]+)')),
    # Qmail with a tri-language intro beginning in spanish
    (_c(r'Your message could not be delivered'),
     _c(r'^-'),
     _c(r'<(?P<addr>[^>]*)>:')),
    # socgen.com
    (_c(r'Your message could not be delivered to'),
     _c(r'^\s*$'),
     _c(r'(?P<addr>[^\s@]+@[^\s@]+)')),
    # dadoservice.it
    (_c(r'Your message has encountered delivery problems'),
     _c(r'Your message reads'),
     _c(r'addressed to\s*(?P<addr>[^\s@]+@[^\s@)]+)')),
    # gomaps.com
    (_c(r'Did not reach the following recipient'),
     _c(r'^\s*$'),
     _c(r'\s(?P<addr>[^\s@]+@[^\s@]+)')),
    # EYOU MTA SYSTEM
    (_c(r'This is the deliver program at'),
     _c(r'^-'),
     _c(r'^(?P<addr>[^\s@]+@[^\s@<>]+)')),
    # A non-standard qmail at ieo.it
    (_c(r'this is the email server at'),
     _c(r'^-'),
     _c(r'\s(?P<addr>[^\s@]+@[^\s@]+)[\s,]')),
    # pla.net.py (MDaemon.PRO ?)
    (_c(r'- no such user here'),
     _c(r'There is no user'),
     _c(r'^(?P<addr>[^\s@]+@[^\s@]+)\s')),
    # mxlogic.net
    (_c(r'The following address failed:'),
     _c(r'Included is a copy of the message header'),
     _c(r'<(?P<addr>[^>]+)>')),
    # fastdnsservers.com
    (_c(r'The following recipient\(s\) could not be reached'),
     _c(r'\s*Error Type'),
     _c(r'^(?P<addr>[^\s@]+@[^\s@<>]+)')),
    # xxx.com (simple_36.txt)
    (_c(r'Could not deliver message to the following recipient'),
     _c(r'\s*-- The header'),
     _c(r'Failed Recipient: (?P<addr>[^\s@]+@[^\s@<>]+)')),
    # mta1.service.uci.edu
    (_c(r'Message not delivered to the following addresses'),
     _c(r'Error detail'),
     _c(r'\s*(?P<addr>[^\s@]+@[^\s@)]+)')),
    # Dovecot LDA Over quota MDN (bogus - should be DSN).
    (_c(r'^Your message'),
     _c(r'^Reporting'),
     _c(r'Your message to <?(?P<addr>[^\s<@]+@[^\s@>]+)>? was automatically'
        ' rejected')),
    # mail.ru
    (_c(r'A message that you sent was rejected'),
     _c(r'This is a copy of your message'),
     _c(r'\s(?P<addr>[^\s@]+@[^\s@]+)')),
    # MailEnable
    (_c(r'Message could not be delivered to some recipients.'),
     _c(r'Message headers follow'),
     _c(r'Recipient: \[SMTP:(?P<addr>[^\s@]+@[^\s@]+)\]')),
    # This one is from Yahoo but dosen't fit the yahoo recognizer format
    (_c(r'wasn\'t able to deliver the following message'),
     _c(r'---Below this line is a copy of the message.'),
     _c(r'To: (?P<addr>[^\s@]+@[^\s@]+)')),
    # From some unknown MTA
    (_c(r'This is a delivery failure notification message'),
     _c(r'The problem appears to be'),
     _c(r'-- (?P<addr>[^\s@]+@[^\s@]+)')),
    # Next one goes here...
    ]


@public
@implementer(IBounceDetector)
class SimpleMatch:
    """Recognizes simple heuristically delimited bounces."""

    PATTERNS = PATTERNS

    def process(self, msg):
        """See `IBounceDetector`."""
        addresses = set()
        new_addresses = set()
        # MAS: This is a mess. The outer loop used to be over the message
        # so we only looped through the message once.  Looping through the
        # message for each set of patterns is obviously way more work, but
        # if we don't do it, problems arise because scre from the wrong
        # pattern set matches first and then acre doesn't match.  The
        # alternative is to split things into separate modules, but then
        # we process the message multiple times anyway.
        for scre, ecre, acre in self.PATTERNS:
            state = ParseState.start
            for line in body_line_iterator(msg):
                if state is ParseState.start:
                    if scre.search(line):
                        state = ParseState.tag_seen
                if state is ParseState.tag_seen:
                    mo = acre.search(line)
                    if mo:
                        address = mo.group('addr')
                        if address:
                            addresses.add(_quopri_decode(address))
                    elif ecre.search(line):
                        break
            if len(addresses) > 0:
                break
        for address in addresses:
            if VALID.match(address):
                new_addresses.add(address)
        return NoTemporaryFailures, new_addresses



import logging

from flufl.bounce._detectors import (
    aol,
    caiwireless,
    dsn,
    exchange,
    exim,
    groupwise,
    llnl,
    microsoft,
    netscape,
    postfix,
    qmail,
    simplematch,
    simplewarning,
    sina,
    smtp32,
    yahoo,
    yale,
    )
from public import public


log = logging.getLogger('flufl.bounce')


def _find_detectors():
    """Get all detectors in order of most specific/authoritative to least"""
    yield from [
        # RFC compliant DSNs.
        dsn.DSN,
        # Detectors for specific providers without heuristics.
        exim.Exim,
        sina.Sina,
        # Provider specific heuristic detectors.
        yahoo.Yahoo,
        yale.Yale,
        smtp32.SMTP32,
        postfix.Postfix,
        qmail.Qmail,
        groupwise.GroupWise,
        microsoft.Microsoft,
        caiwireless.Caiwireless,
        exchange.Exchange,
        netscape.Netscape,
        aol.AOL,
        # Generic and other heuristic detectors.
        simplematch.SimpleMatch,
        simplewarning.SimpleWarning,
        llnl.LLNL,
        ]


@public
def scan_message(msg):
    """Detect the set of all permanently bouncing original recipients.

    :param msg: The bounce message.
    :type msg: `email.message.Message`
    :return: The set of detected original recipients.
    :rtype: set of strings
    """
    permanent_failures = set()
    for detector_class in _find_detectors():
        log.info('Running detector: {}'.format(detector_class))
        try:
            temporary, permanent = detector_class().process(msg)
        except Exception:
            log.exception('Exception in detector: {}'.format(detector_class))
            raise
        permanent_failures.update(permanent)
        if temporary or permanent:
            break
    return permanent_failures


@public
def all_failures(msg):
    """Detect the set of all bouncing original recipients.

    :param msg: The bounce message.
    :type msg: `email.message.Message`
    :return: 2-tuple of the temporary failure set and permanent failure set.
    :rtype: (set of strings, set of string)
    """
    temporary_failures = set()
    permanent_failures = set()
    for detector_class in _find_detectors():
        log.info('Running detector: {}'.format(detector_class))
        temporary, permanent = detector_class().process(msg)
        temporary_failures.update(temporary)
        permanent_failures.update(permanent)
        if temporary or permanent:
            break
    return temporary_failures, permanent_failures



"""Parse bounce messages generated by Exim.

Exim adds an X-Failed-Recipients: header to bounce messages containing
an `addresslist' of failed addresses.
"""

from email.utils import getaddresses
from flufl.bounce.interfaces import IBounceDetector, NoTemporaryFailures
from public import public
from zope.interface import implementer


@public
@implementer(IBounceDetector)
class Exim:
    """Parse bounce messages generated by Exim."""

    def process(self, msg):
        """See `IBounceDetector`."""
        all_failed = msg.get_all('x-failed-recipients', [])
        # all_failed will contain string/unicode values, but the flufl.bounce
        # API requires these to be bytes.  We don't know the encoding, but
        # assume it must be ascii, per the relevant RFCs.
        return (NoTemporaryFailures,
                set(address.encode('us-ascii')
                    for name, address in getaddresses(all_failed)))



"""This appears to be the format for Novell GroupWise and NTMail

X-Mailer: Novell GroupWise Internet Agent 5.5.3.1
X-Mailer: NTMail v4.30.0012
X-Mailer: Internet Mail Service (5.5.2653.19)
"""

import re

from email.message import Message
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from io import BytesIO
from public import public
from zope.interface import implementer


acre = re.compile(b'<(?P<addr>[^>]*)>')


def find_textplain(msg):
    if msg.get_content_type() == 'text/plain':
        return msg
    if msg.is_multipart():
        for part in msg.get_payload():
            if not isinstance(part, Message):
                continue
            ret = find_textplain(part)
            if ret:
                return ret
    return None


@public
@implementer(IBounceDetector)
class GroupWise:
    """Parse Novell GroupWise and NTMail bounces."""

    def process(self, msg):
        """See `IBounceDetector`."""
        if msg.get_content_type() != 'multipart/mixed' or not msg['x-mailer']:
            return NoFailures
        if msg['x-mailer'][:3].lower() not in ('nov', 'ntm', 'int'):
            return NoFailures
        addresses = set()
        # Find the first text/plain part in the message.
        text_plain = find_textplain(msg)
        if text_plain is None:
            return NoFailures
        body = BytesIO(text_plain.get_payload(decode=True))
        for line in body:
            mo = acre.search(line)
            if mo:
                addresses.add(mo.group('addr'))
            elif b'@' in line:
                i = line.find(b' ')
                if i == 0:
                    continue
                if i < 0:
                    addresses.add(line)
                else:
                    addresses.add(line[:i])
        return NoTemporaryFailures, set(addresses)



"""Yahoo! has its own weird format for bounces."""

import re

from email.iterators import body_line_iterator
from email.utils import parseaddr
from enum import Enum
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


tcre = (re.compile(r'message\s+from\s+yahoo\.\S+', re.IGNORECASE),
        re.compile(r'Sorry, we were unable to deliver your message to '
                   r'the following address(\(es\))?\.',
                   re.IGNORECASE),
        )
acre = re.compile(r'<(?P<addr>[^>]*)>:')
ecre = (re.compile(r'--- Original message follows'),
        re.compile(r'--- Below this line is a copy of the message'),
        re.compile(r'---------- Forwarded message ----------'),
        )


class _ParseState(Enum):
    start = 0
    tag_seen = 1
    all_done = 2


@public
@implementer(IBounceDetector)
class Yahoo:
    """Yahoo! bounce detection."""

    def process(self, msg):
        """See `IBounceDetector`."""
        # Yahoo! bounces seem to have a known subject value and something
        # called an x-uidl: header, the value of which seems unimportant.
        sender = parseaddr(msg.get('from', '').lower())[1] or ''
        if not sender.startswith('mailer-daemon@yahoo'):
            return NoFailures
        addresses = set()
        state = _ParseState.start
        for line in body_line_iterator(msg):
            line = line.strip()
            if state is _ParseState.start:
                for cre in tcre:
                    if cre.match(line):
                        state = _ParseState.tag_seen
                        break
            elif state is _ParseState.tag_seen:
                mo = acre.match(line)
                if mo:
                    addresses.add(mo.group('addr').encode('us-ascii'))
                    continue
                for cre in ecre:
                    mo = cre.match(line)
                    if mo:
                        # We're at the end of the error response.
                        state = _ParseState.all_done
                        break
                if state is _ParseState.all_done:
                    break
        return NoTemporaryFailures, addresses



"""Netscape Messaging Server bounce formats.

I've seen at least one NMS server version 3.6 (envy.gmp.usyd.edu.au) bounce
messages of this format.  Bounces come in DSN MIME format, but don't include
any -Recipient: headers.  Gotta just parse the text :(

NMS 4.1 (dfw-smtpin1.email.verio.net) seems even worse, but we'll try to
decipher the format here too.

"""

import re

from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from io import BytesIO
from public import public
from zope.interface import implementer


pcre = re.compile(
    b'This Message was undeliverable due to the following reason:',
    re.IGNORECASE)

acre = re.compile(
    b'(?P<reply>please reply to)?.*<(?P<addr>[^>]*)>',
    re.IGNORECASE)


def flatten(msg, leaves):
    # Give us all the leaf (non-multipart) subparts.
    if msg.is_multipart():
        for part in msg.get_payload():
            flatten(part, leaves)
    else:
        leaves.append(msg)


@public
@implementer(IBounceDetector)
class Netscape:
    """Netscape Messaging Server bounce formats."""

    def process(self, msg):
        """See `IBounceDetector`."""

        # Sigh.  Some NMS 3.6's show
        #     multipart/report; report-type=delivery-status
        # and some show
        #     multipart/mixed;
        if not msg.is_multipart():
            return NoFailures
        # We're looking for a text/plain subpart occuring before a
        # message/delivery-status subpart.
        plainmsg = None
        leaves = []
        flatten(msg, leaves)
        for i, subpart in zip(range(len(leaves)-1), leaves):
            if subpart.get_content_type() == 'text/plain':
                plainmsg = subpart
                break
        if not plainmsg:
            return NoFailures
        # Total guesswork, based on captured examples...
        body = BytesIO(plainmsg.get_payload(decode=True))
        addresses = set()
        for line in body:
            mo = pcre.search(line)
            if mo:
                # We found a bounce section, but I have no idea what the
                # official format inside here is.  :( We'll just search for
                # <addr> strings.
                for line in body:
                    mo = acre.search(line)
                    if mo and not mo.group('reply'):
                        addresses.add(mo.group('addr'))
        return NoTemporaryFailures, addresses



"""Parse bounce messages generated by qmail.

Qmail actually has a standard, called QSBMF (qmail-send bounce message
format), as described in

    http://cr.yp.to/proto/qsbmf.txt

This module should be conformant.

"""

import re

from email.iterators import body_line_iterator
from enum import Enum
from flufl.bounce.interfaces import IBounceDetector, NoTemporaryFailures
from public import public
from zope.interface import implementer


# Other (non-standard?) intros have been observed in the wild.
introtags = [
    "We're sorry. There's a problem",
    'Check your send e-mail address.',
    'Hi. The MTA program at',
    'Hi. This is the',
    'This is the mail delivery agent at',
    'Unfortunately, your mail was not delivered',
    'Your mail message to the following',
    ]
introtags_encoded = [tag.encode('ascii') for tag in introtags]
acre = re.compile(r'<(?P<addr>[^>]*)>:')


class ParseState(Enum):
    start = 0
    intro_paragraph_seen = 1
    recip_paragraph_seen = 2


@public
@implementer(IBounceDetector)
class Qmail:
    """Parse QSBMF format bounces."""

    def process(self, msg):
        """See `IBounceDetector`."""
        addresses = set()
        state = ParseState.start
        for line in body_line_iterator(msg):
            line = line.strip()
            if state is ParseState.start:
                for introtag in (introtags_encoded
                                 if isinstance(line, bytes) else introtags):
                    if line.startswith(introtag):
                        state = ParseState.intro_paragraph_seen
                        break
            elif state is ParseState.intro_paragraph_seen and not line:
                # Looking for the end of the intro paragraph.
                state = ParseState.recip_paragraph_seen
            elif state is ParseState.recip_paragraph_seen:
                if line.startswith('-'):
                    # We're looking at the break paragraph, so we're done.
                    break
                # At this point we know we must be looking at a recipient
                # paragraph.
                mo = acre.match(line)
                if mo:
                    addresses.add(mo.group('addr').encode('us-ascii'))
                # Otherwise, it must be a continuation line, so just ignore it.
            else:
                # We're not looking at anything in particular.
                pass
        return NoTemporaryFailures, addresses



"""Recognizes (some) Microsoft Exchange formats."""

import re

from email.iterators import body_line_iterator
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


scre = re.compile('did not reach the following recipient')
ecre = re.compile('MSEXCH:')
a1cre = re.compile('SMTP=(?P<addr>[^;]+); on ')
a2cre = re.compile('(?P<addr>[^ ]+) on ')


@public
@implementer(IBounceDetector)
class Exchange:
    """Recognizes (some) Microsoft Exchange formats."""

    def process(self, msg):
        """See `IBounceDetector`."""
        addresses = set()
        it = body_line_iterator(msg)
        # Find the start line.
        for line in it:
            if scre.search(line):
                break
        else:
            return NoFailures
        # Search each line until we hit the end line.
        for line in it:
            if ecre.search(line):
                break
            mo = a1cre.search(line)
            if not mo:
                mo = a2cre.search(line)
            if mo:
                # For Python 3 compatibility, the API requires bytes
                address = mo.group('addr').encode('us-ascii')
                addresses.add(address)
        return NoTemporaryFailures, set(addresses)



"""Something which claims
X-Mailer: <SMTP32 vXXXXXX>

What the heck is this thing?  Here's a recent host:

% telnet 207.51.255.218 smtp
Trying 207.51.255.218...
Connected to 207.51.255.218.
Escape character is '^]'.
220 X1 NT-ESMTP Server 208.24.118.205 (IMail 6.00 45595-15)
"""

import re

from email.iterators import body_line_iterator
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


ecre = re.compile('original message follows', re.IGNORECASE)
acre = re.compile(r'''
    (                                             # several different prefixes
    user\ mailbox[^:]*:                           # have been spotted in the
    |delivery\ failed[^:]*:                       # wild...
    |unknown\ user[^:]*:
    |undeliverable\ +to
    |delivery\ userid[^:]*:
    )
    \s*                                           # space separator
    (?P<addr>[^\s]*)                              # and finally, the address
    ''', re.IGNORECASE | re.VERBOSE)


@public
@implementer(IBounceDetector)
class SMTP32:
    """Something which claims

    X-Mailer: <SMTP32 vXXXXXX>
    """

    def process(self, msg):
        mailer = msg.get('x-mailer', '')
        if not mailer.startswith('<SMTP32 v'):
            return NoFailures
        addresses = set()
        for line in body_line_iterator(msg):
            if ecre.search(line):
                break
            mo = acre.search(line)
            if mo:
                addresses.add(mo.group('addr').encode('us-ascii'))
        return NoTemporaryFailures, addresses



"""Yale's mail server is pretty dumb.

Its reports include the end user's name, but not the full domain.  I think we
can usually guess it right anyway.  This is completely based on examination of
the corpse, and is subject to failure whenever Yale even slightly changes
their MTA. :(
"""

import re

from email.utils import getaddresses
from enum import Enum
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from io import BytesIO
from public import public
from zope.interface import implementer


scre = re.compile(b'Message not delivered to the following', re.IGNORECASE)
ecre = re.compile(b'Error Detail', re.IGNORECASE)
acre = re.compile(b'\\s+(?P<addr>\\S+)\\s+')


class ParseState(Enum):
    start = 0
    intro_found = 1


@public
@implementer(IBounceDetector)
class Yale:
    """Parse Yale's bounces (or what used to be)."""

    def process(self, msg):
        """See `IBounceDetector`."""
        if msg.is_multipart():
            return NoFailures
        try:
            whofrom = getaddresses([msg.get('from', '')])[0][1]
            if not whofrom:
                return NoFailures
            username, domain = whofrom.split('@', 1)
        except (IndexError, ValueError):
            return NoFailures
        if username.lower() != 'mailer-daemon':
            return NoFailures
        parts = domain.split('.')
        parts.reverse()
        for part1, part2 in zip(parts, ('edu', 'yale')):
            if part1 != part2:
                return NoFailures
        # Okay, we've established that the bounce came from the mailer-daemon
        # at yale.edu.  Let's look for a name, and then guess the relevant
        # domains.
        names = set()
        body = BytesIO(msg.get_payload(decode=True))
        state = ParseState.start
        for line in body:
            if state is ParseState.start and scre.search(line):
                state = ParseState.intro_found
            elif state is ParseState.intro_found and ecre.search(line):
                break
            elif state is ParseState.intro_found:
                mo = acre.search(line)
                if mo:
                    names.add(mo.group('addr'))
        # Now we have a bunch of names, these are either @yale.edu or
        # @cs.yale.edu.  Add them both.
        addresses = set()
        for name in names:
            addresses.add(name + b'@yale.edu')
            addresses.add(name + b'@cs.yale.edu')
        return NoTemporaryFailures, addresses



"""Recognizes a class of messages from AOL that report only Screen Name."""

import re

from email.utils import parseaddr
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


scre = re.compile(b'mail to the following recipients could not be delivered')


@public
@implementer(IBounceDetector)
class AOL:
    """Recognizes a class of messages from AOL that report only Screen Name."""

    def process(self, msg):
        if msg.get_content_type() != 'text/plain':
            return NoFailures
        if not parseaddr(msg.get('from', ''))[1].lower().endswith('@aol.com'):
            return NoFailures
        addresses = set()
        found = False
        for line in msg.get_payload(decode=True).splitlines():
            if scre.search(line):
                found = True
                continue
            if found:
                local = line.strip()
                if local:
                    if re.search(b'\\s', local):
                        break
                    if b'@' in local:
                        addresses.add(local)
                    else:
                        addresses.add(local + b'@aol.com')
        return NoTemporaryFailures, addresses



import os
import email
import doctest
import logging

from contextlib import closing
from email import message_from_binary_file as parse
from importlib import import_module
from nose2.events import Plugin
from pkg_resources import resource_stream
from unittest import TestCase


DOCTEST_FLAGS = (
    doctest.ELLIPSIS |
    doctest.NORMALIZE_WHITESPACE |
    doctest.REPORT_NDIFF)


def setup(testobj):
    """Test setup."""
    testobj.globs['parse'] = email.message_from_bytes


def initialize(plugin):
    """Initialize logging for the test suite.

    Normally, an application would itself initialize the flufl.bounce logger,
    but when the test suite is run, it is the controlling application.
    Sometimes when you run the test suite, you want additional debugging, so
    you can set the logging level via an environment variable $FLUFL_LOGGING.
    This variable can be a set of semi-colon separated key-value pairs,
    themselves separated by an equal sign.  The keys and values can be
    anything accepted by `logging.basicConfig()`.
    """
    kwargs = {}
    envar = os.environ.get('FLUFL_LOGGING')
    if envar is not None:
        for key_value in envar.split(';'):
            key, equals, value = key_value.partition('=')
            kwargs[key] = value
    logging.basicConfig(**kwargs)


class Detectors(Plugin):
    configSection = 'detectors'

    class DataTest(TestCase):
        def __init__(self, data):
            super().__init__('run_test')
            if len(data) == 3:
                bounce_module, filename, expected = data
                self.is_temporary = False
            else:
                bounce_module, filename, expected, self.is_temporary = data
            self.expected = set(expected)
            self.description = '{}: [{}] detecting {} in {}'.format(
                bounce_module,
                ('T' if self.is_temporary else 'P'),
                self.expected, filename)
            module_name = 'flufl.bounce._detectors.{}'.format(bounce_module)
            module = import_module(module_name)
            with closing(resource_stream(
                    'flufl.bounce.tests.data', filename)) as fp:
                self.message = parse(fp)
            missing = object()
            for name in getattr(module, '__all__', []):
                component_class = getattr(module, name, missing)
                if component_class is missing:
                    raise RuntimeError(
                        'skipping missing __all__ entry: {}'.format(name))
                self.component = component_class()

        def run_test(self):
            # XXX 2011-07-02: We don't currently test temporary failures.
            temporary, permanent = self.component.process(self.message)
            got = (set(temporary) if self.is_temporary else set(permanent))
            self.assertEqual(got, self.expected)

    def loadTestsFromTestCase(self, event):
        if event.testCase.__name__ != 'TestDetectors':
            return
        for data in DATA:
            event.extraTests.append(Detectors.DataTest(data))

    def describeTest(self, event):
        if event.test.__class__ is Detectors.DataTest:
            event.description = event.test.description
            event.handled = True


DATA = (
    # Postfix bounces
    ('postfix', 'postfix_01.txt', [b'xxxxx@local.ie']),
    ('postfix', 'postfix_02.txt', [b'yyyyy@digicool.com']),
    ('postfix', 'postfix_03.txt', [b'ttttt@ggggg.com']),
    ('postfix', 'postfix_04.txt', [b'userx@mail1.example.com']),
    ('postfix', 'postfix_05.txt', [b'userx@example.net']),
    # Exim bounces
    ('exim', 'exim_01.txt', [b'userx@its.example.nl']),
    # SimpleMatch bounces
    ('simplematch', 'sendmail_01.txt', [b'zzzzz@shaft.coal.nl',
                                        b'zzzzz@nfg.nl']),
    ('simplematch', 'simple_01.txt', [b'bbbsss@example.com']),
    ('simplematch', 'simple_02.txt', [b'userx@example.net']),
    ('simplematch', 'simple_04.txt', [b'userx@example.com']),
    ('simplematch', 'newmailru_01.txt', [b'zzzzz@newmail.ru']),
    ('simplematch', 'hotpop_01.txt', [b'userx@example.com']),
    ('simplematch', 'microsoft_03.txt', [b'userx@example.com']),
    ('simplematch', 'simple_05.txt', [b'userx@example.net']),
    ('simplematch', 'simple_06.txt', [b'userx@example.com']),
    ('simplematch', 'simple_07.txt', [b'userx@example.net']),
    ('simplematch', 'simple_08.txt', [b'userx@example.de']),
    ('simplematch', 'simple_09.txt', [b'userx@example.de']),
    ('simplematch', 'simple_10.txt', [b'userx@example.com']),
    ('simplematch', 'simple_11.txt', [b'userx@example.com']),
    ('simplematch', 'simple_12.txt', [b'userx@example.ac.jp']),
    ('simplematch', 'simple_13.txt', [b'userx@example.fr']),
    ('simplematch', 'simple_14.txt', [b'userx@example.com',
                                      b'usery@example.com']),
    ('simplematch', 'simple_15.txt', [b'userx@example.be']),
    ('simplematch', 'simple_16.txt', [b'userx@example.com']),
    ('simplematch', 'simple_17.txt', [b'userx@example.com']),
    ('simplematch', 'simple_18.txt', [b'userx@example.com']),
    ('simplematch', 'simple_19.txt', [b'userx@example.com.ar']),
    ('simplematch', 'simple_20.txt', [b'userx@example.com']),
    ('simplematch', 'simple_23.txt', [b'userx@example.it']),
    ('simplematch', 'simple_24.txt', [b'userx@example.com']),
    ('simplematch', 'simple_25.txt', [b'userx@example.com']),
    ('simplematch', 'simple_26.txt', [b'userx@example.it']),
    ('simplematch', 'simple_27.txt', [b'userx@example.net.py']),
    ('simplematch', 'simple_29.txt', [b'userx@example.com']),
    ('simplematch', 'simple_30.txt', [b'userx@example.com']),
    ('simplematch', 'simple_31.txt', [b'userx@example.fr']),
    ('simplematch', 'simple_32.txt', [b'userx@example.com']),
    ('simplematch', 'simple_33.txt', [b'userx@example.com']),
    ('simplematch', 'simple_34.txt', [b'roland@example.com']),
    ('simplematch', 'simple_36.txt', [b'userx@example.com']),
    ('simplematch', 'simple_37.txt', [b'user@example.edu']),
    ('simplematch', 'simple_38.txt', [b'userx@example.com']),
    ('simplematch', 'simple_39.txt', [b'userx@example.ru']),
    ('simplematch', 'simple_41.txt', [b'userx@example.com']),
    ('simplematch', 'simple_42.txt', []),
    ('simplematch', 'simple_43.txt', []),
    ('simplematch', 'simple_44.txt', [b'user@example.com']),
    ('simplematch', 'bounce_02.txt', [b'userx@example.com']),
    ('simplematch', 'bounce_03.txt', [b'userx@example.uk']),
    ('simplematch', 'yahoo_12.txt', [b'user@yahoo.com']),
    # SimpleWarning
    ('simplewarning', 'simple_03.txt', [b'userx@example.za'], True),
    ('simplewarning', 'simple_21.txt', [b'userx@example.com'], True),
    ('simplewarning', 'simple_22.txt', [b'User@example.org'], True),
    ('simplewarning', 'simple_28.txt', [b'userx@example.com'], True),
    ('simplewarning', 'simple_35.txt', [b'calvin@example.com'], True),
    ('simplewarning', 'simple_40.txt', [b'user@example.com'], True),
    # GroupWise
    ('groupwise', 'groupwise_01.txt', [b'userx@example.EDU']),
    # This one really sucks 'cause it's text/html.  Just make sure it
    # doesn't throw an exception, but we won't get any meaningful
    # addresses back from it.
    ('groupwise', 'groupwise_02.txt', []),
    # Actually, it's from Exchange, and Exchange does recognize it
    ('exchange', 'groupwise_02.txt', [b'userx@example.com']),
    # Not a bounce but has confused groupwise
    ('groupwise', 'groupwise_03.txt', []),
    # Yale's own
    ('yale', 'yale_01.txt', [b'userx@cs.yale.edu',
                             b'userx@yale.edu']),
    # DSN, i.e. RFC 1894
    ('dsn', 'dsn_01.txt', [b'userx@example.com']),
    ('dsn', 'dsn_02.txt', [b'zzzzz@example.uk']),
    ('dsn', 'dsn_03.txt', [b'userx@example.be']),
    ('dsn', 'dsn_04.txt', [b'userx@example.ch']),
    ('dsn', 'dsn_05.txt', [b'userx@example.cz'], True),
    ('dsn', 'dsn_06.txt', [b'userx@example.com'], True),
    ('dsn', 'dsn_07.txt', [b'userx@example.nz'], True),
    ('dsn', 'dsn_08.txt',
     [b'userx@example.de'], True),
    ('dsn', 'dsn_09.txt', [b'userx@example.com']),
    ('dsn', 'dsn_10.txt', [b'anne.person@dom.ain']),
    ('dsn', 'dsn_11.txt', [b'joem@example.com']),
    ('dsn', 'dsn_12.txt', [b'userx@example.jp']),
    ('dsn', 'dsn_13.txt', [b'userx@example.com']),
    ('dsn', 'dsn_14.txt', [b'userx@example.com.dk']),
    ('dsn', 'dsn_15.txt', [b'userx@example.com']),
    ('dsn', 'dsn_16.txt', [b'userx@example.com']),
    ('dsn', 'dsn_17.txt', [b'userx@example.fi'], True),
    ('dsn', 'dsn_18.txt', [b'email@replaced.net']),
    # Microsoft Exchange
    ('exchange', 'microsoft_01.txt', [b'userx@example.COM']),
    ('exchange', 'microsoft_02.txt', [b'userx@example.COM']),
    # Microsoft's `SMTPSVC'?
    ('microsoft', 'microsoft_04.txt', [b'userx@example.COM']),
    # SMTP32
    ('smtp32', 'smtp32_01.txt', [b'userx@example.ph']),
    ('smtp32', 'smtp32_02.txt', [b'userx@example.com']),
    ('smtp32', 'smtp32_03.txt', [b'userx@example.com']),
    ('smtp32', 'smtp32_04.txt', [b'after_another@example.net',
                                 b'one_bad_address@example.net']),
    ('smtp32', 'smtp32_05.txt', [b'userx@example.com']),
    ('smtp32', 'smtp32_06.txt', [b'Absolute_garbage_addr@example.net']),
    ('smtp32', 'smtp32_07.txt', [b'userx@example.com']),
    # Qmail
    ('qmail', 'qmail_01.txt', [b'userx@example.de']),
    ('qmail', 'qmail_02.txt', [b'userx@example.com']),
    ('qmail', 'qmail_03.txt', [b'userx@example.jp']),
    ('qmail', 'qmail_04.txt', [b'userx@example.au']),
    ('qmail', 'qmail_05.txt', [b'userx@example.com']),
    ('qmail', 'qmail_06.txt', [b'ntl@xxx.com']),
    ('qmail', 'qmail_07.txt', [b'user@example.net']),
    ('qmail', 'qmail_08.txt', []),
    # LLNL's custom Sendmail
    ('llnl', 'llnl_01.txt', [b'user1@example.gov']),
    # Netscape's server...
    ('netscape', 'netscape_01.txt', [b'aaaaa@corel.com',
                                     b'bbbbb@corel.com']),
    # Yahoo's proprietary format
    ('yahoo', 'yahoo_01.txt', [b'userx@example.com']),
    ('yahoo', 'yahoo_02.txt', [b'userx@example.es']),
    ('yahoo', 'yahoo_03.txt', [b'userx@example.com']),
    ('yahoo', 'yahoo_04.txt', [b'userx@example.es',
                               b'usery@example.uk']),
    ('yahoo', 'yahoo_05.txt', [b'userx@example.com',
                               b'usery@example.com']),
    ('yahoo', 'yahoo_06.txt', [b'userx@example.com',
                               b'usery@example.com',
                               b'userz@example.com',
                               b'usera@example.com']),
    ('yahoo', 'yahoo_07.txt', [b'userw@example.com',
                               b'userx@example.com',
                               b'usery@example.com',
                               b'userz@example.com']),
    ('yahoo', 'yahoo_08.txt', [b'usera@example.com',
                               b'userb@example.com',
                               b'userc@example.com',
                               b'userd@example.com',
                               b'usere@example.com',
                               b'userf@example.com']),
    ('yahoo', 'yahoo_09.txt', [b'userx@example.com',
                               b'usery@example.com']),
    ('yahoo', 'yahoo_10.txt', [b'userx@example.com',
                               b'usery@example.com',
                               b'userz@example.com']),
    ('yahoo', 'yahoo_11.txt', [b'bad_user@aol.com']),
    ('yahoo', 'yahoo_13.txt', [b'bogus-address@example.net']),
    # sina.com appears to use their own weird SINAEMAIL MTA
    ('sina', 'sina_01.txt', [b'userx@sina.com',
                             b'usery@sina.com']),
    ('aol', 'aol_01.txt', [b'screenname@aol.com']),
    # No address can be detected in these...
    # dumbass_01.txt - We love Microsoft. :(
    # Done
    )



"""Parse bounce messages generated by Postfix.

This also matches something called 'Keftamail' which looks just like Postfix
bounces with the word Postfix scratched out and the word 'Keftamail' written
in in crayon.

It also matches something claiming to be 'The BNS Postfix program', and
'SMTP_Gateway'.  Everybody's gotta be different, huh?
"""

import re

from enum import Enum
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from io import BytesIO
from public import public
from zope.interface import implementer


# Are these heuristics correct or guaranteed?
pcre = re.compile(
    b'[ \\t]*the\\s*(bns)?\\s*(postfix|keftamail|smtp_gateway)',
    re.IGNORECASE)
rcre = re.compile(b'failure reason:$', re.IGNORECASE)
acre = re.compile(b'<(?P<addr>[^>]*)>:')

REPORT_TYPES = ('multipart/mixed', 'multipart/report')


class ParseState(Enum):
    start = 0
    salutation_found = 1


def flatten(msg, leaves):
    # Give us all the leaf (non-multipart) subparts.
    if msg.is_multipart():
        for part in msg.get_payload():
            flatten(part, leaves)
    else:
        leaves.append(msg)


def findaddr(msg):
    addresses = set()
    body = BytesIO(msg.get_payload(decode=True))
    state = ParseState.start
    for line in body:
        # Preserve leading whitespace.
        line = line.rstrip()
        # Yes, use match() to match at beginning of string.
        if state is ParseState.start and (
                pcre.match(line) or rcre.match(line)):
            # Then...
            state = ParseState.salutation_found
        elif state is ParseState.salutation_found and line:
            mo = acre.search(line)
            if mo:
                addresses.add(mo.group('addr'))
            # Probably a continuation line.
    return addresses


@public
@implementer(IBounceDetector)
class Postfix:
    """Parse bounce messages generated by Postfix."""

    def process(self, msg):
        """See `IBounceDetector`."""
        if msg.get_content_type() not in REPORT_TYPES:
            return NoFailures
        # We're looking for the plain/text subpart with a Content-Description:
        # of 'notification'.
        leaves = []
        flatten(msg, leaves)
        for subpart in leaves:
            content_type = subpart.get_content_type()
            content_desc = subpart.get('content-description', '').lower()
            if content_type == 'text/plain' and content_desc == 'notification':
                return NoTemporaryFailures, set(findaddr(subpart))
        return NoFailures



"""Microsoft's `SMTPSVC' nears I kin tell."""

import re

from enum import Enum
from flufl.bounce._detectors.simplematch import VALID
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from io import BytesIO
from public import public
from zope.interface import implementer


scre = re.compile(br'transcript of session follows', re.IGNORECASE)


class ParseState(Enum):
    start = 0
    tag_seen = 1


@public
@implementer(IBounceDetector)
class Microsoft:
    """Microsoft's `SMTPSVC' nears I kin tell."""

    def process(self, msg):
        if msg.get_content_type() != 'multipart/mixed':
            return NoFailures
        # Find the first subpart, which has no MIME type.
        try:
            subpart = msg.get_payload(0)
        except IndexError:
            # The message *looked* like a multipart but wasn't.
            return NoFailures
        data = subpart.get_payload(decode=True)
        if isinstance(data, list):
            # The message is a multi-multipart, so not a matching bounce.
            return NoFailures
        body = BytesIO(data)
        state = ParseState.start
        addresses = set()
        for line in body:
            if state is ParseState.start:
                if scre.search(line):
                    state = ParseState.tag_seen
            elif state is ParseState.tag_seen:
                if VALID.match(line.strip()):
                    addresses.add(line.strip())
        return NoTemporaryFailures, set(addresses)



"""LLNL's custom Sendmail bounce message."""

import re

from email.iterators import body_line_iterator
from flufl.bounce.interfaces import (
    IBounceDetector, NoFailures, NoTemporaryFailures)
from public import public
from zope.interface import implementer


acre = re.compile(r',\s*(?P<addr>\S+@[^,]+),', re.IGNORECASE)


@public
@implementer(IBounceDetector)
class LLNL:
    """LLNL's custom Sendmail bounce message."""

    def process(self, msg):
        """See `IBounceDetector`."""

        for line in body_line_iterator(msg):
            mo = acre.search(line)
            if mo:
                address = mo.group('addr').encode('us-ascii')
                return NoTemporaryFailures, set([address])
        return NoFailures



"""Test the bounce detection modules."""

import unittest

from contextlib import closing
from email import message_from_binary_file as parse, message_from_string
from flufl.bounce._detectors.caiwireless import Caiwireless
from flufl.bounce._detectors.microsoft import Microsoft
from flufl.bounce._detectors.smtp32 import SMTP32
from flufl.bounce._scan import all_failures, scan_message
from pkg_resources import resource_stream


COMMASPACE = b', '


class TestOtherBounces(unittest.TestCase):
    def test_SMTP32_failure(self):
        # This file has no X-Mailer: header
        with closing(resource_stream('flufl.bounce.tests.data',
                                     'postfix_01.txt')) as fp:
            msg = parse(fp)
        self.failIf(msg['x-mailer'] is not None)
        temporary, permanent = SMTP32().process(msg)
        self.failIf(temporary)
        self.failIf(permanent)

    def test_caiwireless(self):
        # BAW: this is a mostly bogus test; I lost the samples. :(
        msg = message_from_string("""\
Content-Type: multipart/report; boundary=BOUNDARY

--BOUNDARY

--BOUNDARY--

""")
        temporary, permanent = Caiwireless().process(msg)
        self.failIf(temporary)
        self.failIf(permanent)

    def test_microsoft(self):
        # BAW: similarly as above, I lost the samples. :(
        msg = message_from_string("""\
Content-Type: multipart/report; boundary=BOUNDARY

--BOUNDARY

--BOUNDARY--

""")
        temporary, permanent = Microsoft().process(msg)
        self.failIf(temporary)
        self.failIf(permanent)

    def test_caiwireless_lp_917720(self):
        # https://bugs.launchpad.net/flufl.bounce/+bug/917720
        with closing(resource_stream('flufl.bounce.tests.data',
                                     'simple_01.txt')) as fp:
            msg = parse(fp)
        self.assertEqual(scan_message(msg), set([b'bbbsss@example.com']))


class TestScanAllDetectors(unittest.TestCase):
    # Ensure that _scan runs detectors.
    def test_llnl(self):
        with closing(resource_stream('flufl.bounce.tests.data',
                                     'llnl_01.txt')) as fp:
            msg = parse(fp)
        self.assertEqual(scan_message(msg), set([b'user1@example.gov']))
        temporary, permanent = all_failures(msg)
        self.assertEqual(temporary, set())
        self.assertEqual(permanent, set([b'user1@example.gov']))

    def test_unrecognized(self):
        with closing(resource_stream('flufl.bounce.tests.data',
                                     'dumbass_01.txt')) as fp:
            msg = parse(fp)
        self.assertEqual(scan_message(msg), set())
        temporary, permanent = all_failures(msg)
        self.assertEqual(temporary, set())
        self.assertEqual(permanent, set())

    def test_warning(self):
        with closing(resource_stream('flufl.bounce.tests.data',
                                     'simple_03.txt')) as fp:
            msg = parse(fp)
        self.assertEqual(scan_message(msg), set())
        temporary, permanent = all_failures(msg)
        self.assertEqual(temporary, set([b'userx@example.za']))
        self.assertEqual(permanent, set())


class TestDetectors(unittest.TestCase):
    # This is a pure placeholder for the nose2 plugin in
    # flufl.bounce.testing.helpers.Detectors.
    pass



from setup_helpers import description, get_version, require_python
from setuptools import setup, find_packages


require_python(0x30400f0)
__version__ = get_version('flufl/bounce/__init__.py')


setup(
    name='flufl.bounce',
    version=__version__,
    namespace_packages=['flufl'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    maintainer='Barry Warsaw',
    maintainer_email='barry@python.org',
    description=description('README.rst'),
    license='ASLv2',
    url='https://fluflbounce.readthedocs.io/en/latest/',
    download_url='https://pypi.python.org/pypi/flufl.bounce',
    install_requires=[
        'atpublic',
        'zope.interface',
        ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Communications :: Email',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )



"""Parse RFC 3464 (i.e. DSN) bounce formats.

RFC 3464 obsoletes 1894 which was the old DSN standard.  This module has not
been audited for differences between the two.
"""

from email.iterators import typed_subpart_iterator
from email.utils import parseaddr
from flufl.bounce.interfaces import IBounceDetector
from public import public
from zope.interface import implementer


@public
@implementer(IBounceDetector)
class DSN:
    """Parse RFC 3464 (i.e. DSN) bounce formats."""

    def process(self, msg):
        """See `IBounceDetector`."""
        # Iterate over each message/delivery-status subpart.
        failed_addresses = []
        delayed_addresses = []
        for part in typed_subpart_iterator(msg, 'message', 'delivery-status'):
            if not part.is_multipart():
                # Huh?
                continue
            # Each message/delivery-status contains a list of Message objects
            # which are the header blocks.  Iterate over those too.
            for msgblock in part.get_payload():
                address_set = None
                # We try to dig out the Original-Recipient (which is optional)
                # and Final-Recipient (which is mandatory, but may not exactly
                # match an address on our list).  Some MTA's also use
                # X-Actual-Recipient as a synonym for Original-Recipient, but
                # some apparently use that for other purposes :(
                #
                # Also grok out Action so we can do something with that too.
                action = msgblock.get('action', '').lower()
                # Some MTAs have been observed that put comments on the action.
                if action.startswith('delayed'):
                    address_set = delayed_addresses
                # opensmtpd uses non-compliant Action: error
                elif action.startswith('fail') or action.startswith('error'):
                    address_set = failed_addresses
                else:
                    # Some non-permanent failure, so ignore this block.
                    continue
                params = []
                foundp = False
                for header in ('original-recipient', 'final-recipient'):
                    for k, v in msgblock.get_params([], header):
                        if k.lower() == 'rfc822':
                            foundp = True
                        else:
                            params.append(k)
                    if foundp:
                        # Note that params should already be unquoted.
                        address_set.extend(params)
                        break
                    else:
                        # MAS: This is a kludge, but
                        # SMTP-GATEWAY01.intra.home.dk has a final-recipient
                        # with an angle-addr and no address-type parameter at
                        # all. Non-compliant, but ...
                        for param in params:
                            if param.startswith('<') and param.endswith('>'):
                                address_set.append(param[1:-1])
        # There may be Nones in the current set of failures, so filter those
        # out of both sets.  Also, for Python 3 compatibility, the API
        # requires byte addresses.
        return (
            # First, the delayed, or temporary failures.
            set(parseaddr(address)[1].encode('us-ascii')
                for address in delayed_addresses
                if address is not None),
            # And now the failed or permanent failures.
            set(parseaddr(address)[1].encode('us-ascii')
                for address in failed_addresses
                if address is not None)
            )



"""Recognizes simple heuristically delimited warnings."""

from flufl.bounce._detectors.simplematch import SimpleMatch, _c
from flufl.bounce.interfaces import NoPermanentFailures
from public import public


# This is a list of tuples of the form
#
#     (start cre, end cre, address cre)
#
# where 'cre' means compiled regular expression, start is the line just before
# the bouncing address block, end is the line just after the bouncing address
# block, and address cre is the regexp that will recognize the addresses.  It
# must have a group called 'addr' which will contain exactly and only the
# address that bounced.
PATTERNS = [
    # pop3.pta.lia.net
    (_c('The address to which the message has not yet been delivered is'),
     _c('No action is required on your part'),
     _c(r'\s*(?P<addr>\S+@\S+)\s*')),
    # MessageSwitch
    (_c('Your message to:'),
     _c('This is just a warning, you do not need to take any action'),
     _c(r'\s*(?P<addr>\S+@\S+)\s*')),
    # Symantec_AntiVirus_for_SMTP_Gateways
    (_c('Your message with Subject:'),
     _c('Delivery attempts will continue to be made'),
     _c(r'\s*(?P<addr>\S+@\S+)\s*')),
    # googlemail.com warning
    (_c('Delivery to the following recipient has been delayed'),
     _c('Message will be retried'),
     _c(r'\s*(?P<addr>\S+@\S+)\s*')),
    # Exchange warning message.
    (_c('This is an advisory-only email'),
     _c('has been postponed'),
     _c('"(?P<addr>[^"]+)"')),
    # kundenserver.de
    (_c('not yet been delivered'),
     _c('No action is required on your part'),
     _c(r'\s*<?(?P<addr>\S+@[^>\s]+)>?\s*')),
    # Next one goes here...
    ]


@public
class SimpleWarning(SimpleMatch):
    """Recognizes simple heuristically delimited warnings."""

    PATTERNS = PATTERNS

    def process(self, msg):
        """See `SimpleMatch`."""
        # Since these are warnings, they're classified as temporary failures.
        # There are no permanent failures.
        (temporary,
         permanent_really_temporary) = super(SimpleWarning, self).process(msg)
        return permanent_really_temporary, NoPermanentFailures
